import typing

import numpy as np
import numpy.typing as npt
import torch
from torch import nn

from helios import core

from .functional import (
    calculate_mae,
    calculate_mae_torch,
    calculate_mAP,
    calculate_psnr,
    calculate_psnr_torch,
    calculate_ssim,
    calculate_ssim_torch,
)

METRICS_REGISTRY = core.Registry("metrics")
"""
Global instance of the registry for metric functions.

Example:
    .. code-block:: python

        import helios.metrics as hlm

        # This automatically registers your metric function.
        @hlm.METRICS_REGISTRY.register
        class MyMetric:
            ...

        # Alternatively you can manually register a metric function like this:
        hlm.METRICS_REGISTRY.register(MyMetric)
"""


def create_metric(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> nn.Module:
    """
    Create the metric function for the given type.

    Args:
        type_name: the type of the loss to create.
        args: positional arguments to pass into the metric.
        kwargs: keyword arguments to pass into the metric.

    Returns:
        The metric function
    """
    return METRICS_REGISTRY.get(type_name)(*args, **kwargs)


@METRICS_REGISTRY.register
class CalculatePSNR(nn.Module):
    """
    Calculate PSNR (Peak Signal-to-Noise Ratio).

    Implementation follows: `<https://en.wikipedia.org/wiki/Peak_signal-to-noise_ratio>`__.
    Note that the input_order is only needed if you plan to evaluate Numpy images. It can
    be left as default otherwise.

    Args:
        crop_border: Cropped pixels in each edge of an image. These pixels are not
            involved in the calculation.
        input_order: Whether the input order is "HWC" or "CHW". Defaults to "HWC".
        test_y_channel: Test on Y channel of YCbCr. Defaults to false.
    """

    def __init__(
        self, crop_border: int, input_order: str = "HWC", test_y_channel: bool = False
    ):
        """Construct the PSNR metric."""
        super().__init__()

        self._crop_border = crop_border
        self._input_order = input_order
        self._test_y_channel = test_y_channel

    def forward(
        self, img: npt.NDArray | torch.Tensor, img2: npt.NDArray | torch.Tensor
    ) -> float:
        """
        Calculate the PSNR metric.

        Args:
            img: Images with range :math:`[0, 255]`.
            img2: Images with range :math:`[0, 255]`.

        Returns:
            PSNR value.
        """
        if isinstance(img, torch.Tensor) and isinstance(img2, torch.Tensor):
            return calculate_psnr_torch(
                img, img2, self._crop_border, self._test_y_channel
            )
        assert isinstance(img, np.ndarray) and isinstance(img2, np.ndarray)
        return calculate_psnr(
            img, img2, self._crop_border, self._input_order, self._test_y_channel
        )


@METRICS_REGISTRY.register
class CalculateSSIM(nn.Module):
    """
    Calculate SSIM (structural similarity).

    Implementation follows: 'Image quality assesment: From error visibility to structural
    similarity'. Results are identical to those of the official MATLAB code in
    `<https://ece.uwaterloo.ca/~z70wang/research/ssim/>`__.
    For three-channel images, SSIM is calculated for each channel and then
    averaged.

    Args:
        crop_border: Cropped pixels in each edge of an image. These pixels are not
            involved in the calculation.
        input_order: Whether the input order is "HWC" or "CHW". Defaults to "HWC".
        test_y_channel: Test on Y channel of YCbCr. Defaults to false.
    """

    def __init__(
        self, crop_border: int, input_order: str = "HWC", test_y_channel: bool = False
    ):
        """Construct the SSIM metric."""
        super().__init__()

        self._crop_border = crop_border
        self._input_order = input_order
        self._test_y_channel = test_y_channel

    def forward(
        self, img: npt.NDArray | torch.Tensor, img2: npt.NDArray | torch.Tensor
    ) -> float:
        """
        Calculate the SSIM metric.

        Args:
            img: Images with range :math:`[0, 255]`.
            img2: Images with range :math:`[0, 255]`.

        Returns:
            PSNR value.
        """
        if isinstance(img, torch.Tensor) and isinstance(img2, torch.Tensor):
            return calculate_ssim_torch(
                img, img2, self._crop_border, self._test_y_channel
            )
        assert isinstance(img, np.ndarray) and isinstance(img2, np.ndarray)
        return calculate_ssim(
            img, img2, self._crop_border, self._input_order, self._test_y_channel
        )


@METRICS_REGISTRY.register
class CalculateMAP(nn.Module):
    """
    Calculate the mAP (Mean Average Precision).

    Implementation follows:
    `<https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Mean_average_precision>`__.
    """

    def forward(self, targs: npt.NDArray, preds: npt.NDArray) -> float:
        """
        Calculate the mAP (Mean Average Precision).

        Args:
            targs: target (inferred) labels in range :math:`[0, 1]`.
            preds: predicate labels in range :math:`[0, 1]`.

        Returns:
            The mAP score
        """
        return calculate_mAP(targs, preds)


@METRICS_REGISTRY.register
class CalculateMAE(nn.Module):
    """
    Compute the MAE (Mean-Average Precision) score.

    Implementation follows: `<https://en.wikipedia.org/wiki/Mean_absolute_error>`__.
    The scale argument is used in the event that the input arrays are not in the range
    :math:`[0, 1]` but instead have been scaled to be in the range :math:`[0, N]` where
    :math:`N` is the factor. For example, if the arrays are images in the range
    :math:`[0, 255]`, then the scaling factor should be set to 255. If the arrays are
    already in the range :math:`[0, 1]`, then the scale can be omitted.

    Args:
        scale: scaling factor that was used on the input tensors. Defaults to 1.
    """

    def __init__(self, scale: float = 1):
        """Construct the MAE metric."""
        super().__init__()
        self._scale = scale

    def forward(
        self, pred: npt.NDArray | torch.Tensor, gt: npt.NDArray | torch.Tensor
    ) -> float:
        """
        Calculate the MAE metric.

        Args:
            pred: predicate (inferred) data.
            gt: ground-truth data.

        Returns:
            The MAE score
        """
        if isinstance(pred, torch.Tensor) and isinstance(gt, torch.Tensor):
            return calculate_mae_torch(pred, gt, self._scale)
        assert isinstance(pred, np.ndarray) and isinstance(gt, np.ndarray)
        return calculate_mae(pred, gt, self._scale)



from __future__ import annotations

import abc
import dataclasses as dc
import typing

import torch

from helios import core

if typing.TYPE_CHECKING:
    from ..model import Model
    from ..trainer import Trainer, TrainingState

# ruff: noqa: B027


@dc.dataclass
class UniquePluginOverrides:
    """
    Set of flags that determine the unique overrides a plug-in can have.

    In order to avoid conflicts, two plug-ins should *not* be able to perform the same
    action twice. For example, it shouldn't be possible to have two distinct plug-ins
    perform processing on the training batch as that would cause undefined behaviour. This
    structure therefore holds all the possible overrides a plug-in might have that
    **must** remain unique.

    Args:
        training_batch: if true, the plug-in performs processing on the training batch.
        validation_batch: if true, the plug-in performs processing on the validation
            batch.
        testing_batch: if true, the plug-in performs processing on the testing batch.
        should_training_stop: if true, the plug-in can arbitrarily stop training.
    """

    training_batch: bool = False
    validation_batch: bool = False
    testing_batch: bool = False
    should_training_stop: bool = False


PLUGIN_REGISTRY = core.Registry("plug-in")
"""
Global instance of the registry for plug-ins.

By default, the registry contains the following plug-ins:

.. list-table:: Schedulers
    :header-rows: 1

    * - Plugin
      - Name
    * - :py:class:`helios.plugins.plugin.CUDAPlugin`
      - CUDAPlugin
    * - :py:class:`helios.plugins.optuna.OptunaPlugin`
      - OptunaPlugin

.. note::
    The :py:class:`~helios.plugins.optuna.OptunaPlugin` is only registered if the module
    is imported somewhere in the code. Otherwise it won't be registered.

Example:
    .. code-block:: python

        import helios.plug-ins as hlp

        # This automatically registers your plug-in
        @hlp.PLUGIN_REGISTRY
        class MyPlugin(hlp.Plugin):
            ...

        # Alternatively, you can manually register a plug-in like this:
        hlp.PLUGIN_REGISTRY.register(MyPlugin)
"""


def create_plugin(type_name: str, *args: typing.Any, **kwargs: typing.Any) -> Plugin:
    """
    Create the plug-in for the given type.

    Args:
        type_name: the type of the plug-in to create.
        args: positional arguments to pass into the plug-in.
        kwargs: keyword arguments to pass into the plug-in.

    Returns:
        The plug-in.
    """
    return PLUGIN_REGISTRY.get(type_name)(*args, **kwargs)


class Plugin(abc.ABC):
    """
    Base class for plug-ins that extend the functionality of the Helios trainer.

    You can use this class to customize the behaviour of training to achieve a variety of
    objectives.
    The plug-ins have a similar API to the :py:class:`~helios.model.model.Model` class.
    The only major difference is that the plug-in functions are called *before* the
    corresponding model functions, providing the ability to override the model if
    necessary.

    Args:
        plug_id: the string with which the plug-in will be registered in the trainer
            plug-in table.
    """

    def __init__(self, plug_id: str):
        """Create the plug-in."""
        self._trainer: Trainer | None
        self._plug_id = plug_id

        self._is_distributed: bool = False
        self._map_loc: str | dict[str, str] = ""
        self._device: torch.device | None = None
        self._rank: int = 0
        self._overrides = UniquePluginOverrides()

    @property
    def unique_overrides(self) -> UniquePluginOverrides:
        """The set of unique overrides the plug-in uses."""
        return self._overrides

    @property
    def is_distributed(self) -> bool:
        """Flag controlling whether distributed training is being used or not."""
        return self._is_distributed

    @is_distributed.setter
    def is_distributed(self, val: bool) -> None:
        self._is_distributed = val

    @property
    def map_loc(self) -> str | dict[str, str]:
        """The location to map loaded weights from a checkpoint or pre-trained file."""
        return self._map_loc

    @map_loc.setter
    def map_loc(self, loc: str | dict[str, str]) -> None:
        self._map_loc = loc

    @property
    def device(self) -> torch.device:
        """The device on which the plug-in is running."""
        return core.get_from_optional(self._device)

    @device.setter
    def device(self, dev: torch.device) -> None:
        self._device = dev

    @property
    def rank(self) -> int:
        """The local rank (device id) that the plug-in is running on."""
        return self._rank

    @rank.setter
    def rank(self, r: int) -> None:
        self._rank = r

    @property
    def trainer(self) -> Trainer:
        """Reference to the trainer."""
        return core.get_from_optional(self._trainer)

    @trainer.setter
    def trainer(self, t: Trainer) -> None:
        self._trainer = t

    def configure_trainer(self, trainer: Trainer) -> None:
        """
        Configure the trainer before training or testing.

        This function can be used to set certain properties of the trainer. For example,
        it can be used to assign valid exceptions that the plug-in requires or to register
        the plug-in itself in the trainer.

        Args:
            trainer: the trainer instance.
        """

    def configure_model(self, model: Model) -> None:
        """
        Configure the model before training or testing.

        This function can be used to set certain properties of the model. For example, it
        can be used to override the save name of the model.
        """

    @abc.abstractmethod
    def setup(self) -> None:
        """Construct all required state for the plug-in."""

    def on_training_start(self) -> None:
        """Perform any necessary actions when training starts."""

    def process_training_batch(
        self, batch: typing.Any, state: TrainingState
    ) -> typing.Any:
        """
        Process the training batch.

        This function can be used to perform any processing on the training batch *prior*
        to the call to :py:meth:`~helios.model.model.Model.train_step`. For example, this
        can be used to filter out elements in a batch to reduce its size, or it can be
        used to move all elements in the batch to a set device.

        Args:
            batch: the batch data returned from the dataset.
            state: the current training state.
        """
        return batch

    def on_training_end(self) -> None:
        """Perform any necessary actions when training ends."""

    def on_validation_start(self, validation_cycle: int) -> None:
        """
        Perform any necessary actions when validation starts.

        Args:
            validation_cycle: the validation cycle number.
        """

    def process_validation_batch(self, batch: typing.Any, step: int) -> typing.Any:
        """
        Process the validation batch.

        This function can be used to perform any processing on the validation batch
        *prior* to the call to :py:meth:`~helios.model.model.Model.valid_step`. For
        example, this can be used to filter out elements in a batch to reduce its size,
        or it can be used to move all elements in the batch to a set device.

        Args:
            batch: the batch data returned from the dataset.
            step: the current validation batch.
        """
        return batch

    def on_validation_end(self, validation_cycle: int) -> None:
        """
        Perform any necessary actions when validation ends.

        Args:
            validation_cycle: the validation cycle number
        """

    def should_training_stop(self) -> bool:
        """
        Determine whether training should stop or continue.

        Returns:
            False if training should continue, true otherwise.
        """
        return False

    def load_state_dict(self, state_dict: dict[str, typing.Any]) -> None:
        """
        Load the plug-in state from the given state dictionary.

        Use this function to restore any state from a checkpoint.

        Args:
            state_dict: the state dictionary to load from.
        """

    def state_dict(self) -> dict[str, typing.Any]:
        """
        Get the state dictionary of the plug-in.

        Use this function to save any state that you require for checkpoints.

        Returns:
            The state dictionary of the plug-in.
        """
        return {}

    def on_testing_start(self) -> None:
        """Perform any actions when testing starts."""

    def process_testing_batch(self, batch: typing.Any, step: int) -> typing.Any:
        """
        Process the testing batch.

        This function can be used to perform any processing on the testing batch
        *prior* to the call to :py:meth:`~helios.model.model.Model.test_step`. For
        example, this can be used to filter out elements in a batch to reduce its size,
        or it can be used to move all elements in the batch to a set device.

        Args:
            batch: the batch data returned from the dataset.
            step: the current testing batch number.
        """
        return batch

    def on_testing_end(self) -> None:
        """Perform any necessary actions when testing ends."""

    def _append_train_exceptions(
        self, exc: type[Exception] | list[type[Exception]], trainer: Trainer
    ) -> None:
        """
        Append exception type(s) to the list of valid train exceptions.

        Args:
            exc: valid exception type or a list of valid exception types.
            trainer: the trainer instance.
        """
        exc = core.convert_to_list(exc)  # type: ignore[arg-type]
        trainer.train_exceptions.extend(exc)  # type: ignore[arg-type]

    def _append_test_exceptions(
        self, exc: type[Exception] | list[type[Exception]], trainer: Trainer
    ) -> None:
        """
        Append exception type(s) to the list of valid test exceptions.

        Args:
            exc: valid exception type or a list of valid exception types.
            trainer: the trainer instance.
        """
        exc = core.convert_to_list(exc)  # type: ignore[arg-type]
        trainer.test_exceptions.extend(exc)  # type: ignore[arg-type]

    def _register_in_trainer(self, trainer: Trainer) -> None:
        """
        Register the plug-in instance in the trainer table.

        The plug-in will be registered with the value provided to ``plug_id``. If another
        plug-in is found with the same ID, an exception is raised.

        Args:
            trainer: the trainer instance.

        Raises:
            KeyError: if another plug-in with the same ID is found.
        """
        if self._plug_id in trainer.plugins:
            raise KeyError(
                f"error: a plug-in with id {self._plug_id} has already been registered"
            )

        trainer.plugins[self._plug_id] = self


@PLUGIN_REGISTRY.register
class CUDAPlugin(Plugin):
    """
    Plug-in to move elements of a training batch to a GPU.

    This plug-in can be used to move the elements of a training batch to the currently
    selected device automatically *prior* to the call to
    :py:meth:`~helios.model.model.Model.train_step`. The device is automatically assigned
    by the :py:class:`helios.trainer.Trainer` when training or testing starts.

    In order to cover the largest possible number of structures, the plug-in can handle
    the following containers:

    #. Single tensors
    #. Lists. Note that the elements of the list need not all be tensors. If any
       tensors are present, they are automatically moved to the device.
    #. Dictionaries. Similar to the list, not all the elements of the dictionary have
       to be tensors. Any tensors are detected automatically.

    .. warning::
        The plug-in is **not** designed to work with nested structures. In other words, if
        a list of dictionaries is passed in, the plug-in **will not** recognise any
        tensors contained inside the dictionary. Similarly, if a dictionary contains
        nested dictionaries (or any other container), the plug-in won't recognise them.

    .. warning::
        The use of this plug-in **requires** CUDA being enabled. If CUDA is not present,
        an exception is raised.

    .. note::
        If you require custom handling for your specific data types, you can override the
        behaviour of the plug-in by deriving from it. See the example below for details.

        Example:
            .. code-block:: python

                import helios.plug-ins as hlp

                class MyCUDAPlugin(hlp.CUDAPlugin):
                    def _move_collection_to_device(self, batch: <your-type>):
                        # Suppose our batch is a list:
                        for i in range(len(batch)):
                            batch[i] = batch[i].to(self.device)
    """

    plugin_id = "cuda"

    def __init__(self):
        """Create the plug-in."""
        super().__init__(self.plugin_id)
        core.cuda.requires_cuda_support()

        # Ensure that no-one else can manipulate the batch.
        self._overrides.training_batch = True
        self._overrides.validation_batch = True
        self._overrides.testing_batch = True

    def setup(self) -> None:
        """No-op setup function."""

    def configure_trainer(self, trainer: Trainer) -> None:
        """
        Register the plug-in instance into the trainer.

        The plug-in will be registered under the name ``cuda``.

        Args:
            trainer: the trainer instance.
        """
        self._register_in_trainer(trainer)

    def _move_collection_to_device(
        self, batch: torch.Tensor | list | dict | tuple
    ) -> torch.Tensor | list | dict | tuple:
        if isinstance(batch, torch.Tensor):
            batch = batch.to(self.device)
        elif isinstance(batch, list):
            for i in range(len(batch)):
                if isinstance(batch[i], torch.Tensor):
                    batch[i] = batch[i].to(self.device)
        elif isinstance(batch, tuple):
            as_list = list(batch)
            for i in range(len(as_list)):
                if isinstance(as_list[i], torch.Tensor):
                    as_list[i] = as_list[i].to(self.device)

            batch = tuple(as_list)
        elif isinstance(batch, dict):
            for key in batch:
                if isinstance(batch[key], torch.Tensor):
                    batch[key] = batch[key].to(self.device)
        else:
            raise RuntimeError(f"error: batch has unknown type {type(batch)}")

        return batch

    def process_training_batch(
        self, batch: typing.Any, state: TrainingState
    ) -> typing.Any:
        """
        Move the training batch to the GPU.

        Args:
            batch: the batch returned by the training dataset.
            state: the current training state.
        """
        return self._move_collection_to_device(batch)

    def process_validation_batch(self, batch: typing.Any, step: int) -> typing.Any:
        """
        Move the validation batch to the GPU.

        Args:
            batch: the batch returned by the training dataset.
            state: the current training state.
        """
        return self._move_collection_to_device(batch)

    def process_testing_batch(self, batch: typing.Any, step: int) -> typing.Any:
        """
        Move the testing batch to the GPU.

        Args:
            batch: the batch returned by the training dataset.
            state: the current training state.
        """
        return self._move_collection_to_device(batch)



"""
Metrics package for Helios.

This contains the classes, utilities, and registries related to metric functions.
"""

from .metrics import (
    METRICS_REGISTRY,
    CalculateMAE,
    CalculateMAP,
    CalculatePSNR,
    CalculateSSIM,
    create_metric,
)

__all__ = [
    "METRICS_REGISTRY",
    "CalculateMAE",
    "CalculateMAP",
    "CalculatePSNR",
    "CalculateSSIM",
    "create_metric",
]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "Helios"
copyright = "2024, Mauricio A. Rovira Galvez"  # noqa: A001
author = "Mauricio A. Rovira Galvez"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.napoleon",
    "autoapi.extension",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
]
autosummary_generate = True
templates_path = ["_templates"]
exclude_patterns = []  # type: ignore[var-annotated]

autoapi_dirs = ["../../src"]
autoapi_options = [
    "members",
    "undoc-members",
    "show-inheritance",
    "show-module-summary",
    "special-members",
]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_theme_options = {
    "sidebar_hide_name": True,
}
html_static_path = ["_static"]
html_logo = "../../data/logo/logo-transparent.png"



"""
Plugin package for Helios.

This contains the classes, utilities, and registries related to plugins.
"""

from .plugin import (
    PLUGIN_REGISTRY,
    CUDAPlugin,
    Plugin,
    UniquePluginOverrides,
    create_plugin,
)

__all__ = [
    "PLUGIN_REGISTRY",
    "CUDAPlugin",
    "Plugin",
    "UniquePluginOverrides",
    "create_plugin",
]



import argparse
import pathlib
import platform

import torch
import tqdm

from helios import core

from ._version import __version__
from .trainer import TrainingState, register_trainer_types_for_safe_load


def migrate_checkpoints_to_current_version(root: pathlib.Path) -> None:
    """
    Migrate existing checkpoints from a previous version of Helios to the current version.

    This function exists to provide backwards compatibility with checkpoints produced by
    older versions of Helios.

    Args:
        root: the root where the checkpoints are stored.
    """
    register_trainer_types_for_safe_load()

    for chkpt_path in tqdm.tqdm(
        list(root.glob("*.pth")), desc="Migrating checkpoints", unit="chkpt"
    ):
        state = core.safe_torch_load(chkpt_path)

        # Pre-v1.0, checkpoints didn't have a version key.
        if "version" not in state:
            state["version"] = __version__

        # Pre-v1.1, the TrainingState struct is saved as a dictionary, not the object
        # itself.
        if isinstance(state["training_state"], dict):
            state["training_state"] = TrainingState(**state["training_state"])

        torch.save(state, chkpt_path)


def _main() -> None:
    parser = argparse.ArgumentParser(
        description="Migration tool to convert checkpoints generated by versions of "
        f"Helios prior to {__version__}"
    )
    parser.add_argument(
        "root",
        metavar="ROOT",
        nargs=1,
        type=str,
        help="Root where the checkpoints are stored",
    )

    args = parser.parse_args()
    root = args.root[0]

    # Temporarily re-direct PosixPath to WindowsPath on Windows to avoid problems.
    tmp: type[pathlib.WindowsPath] | type[pathlib.PosixPath] | None = None
    if platform.system() == "Windows":
        tmp = pathlib.PosixPath
        pathlib.PosixPath = pathlib.WindowsPath  # type: ignore[assignment, misc]
    elif platform.system() == "Linux":
        tmp = pathlib.WindowsPath
        pathlib.WindowsPath = pathlib.PosixPath  # type: ignore[assignment, misc]

    migrate_checkpoints_to_current_version(pathlib.Path(root))

    # Restore to default values.
    if platform.system() == "Windows":
        pathlib.PosixPath = tmp  # type: ignore[assignment, misc]
    elif platform.system() == "Linux":
        pathlib.WindowsPath = tmp  # type: ignore[assignment, misc]


if __name__ == "__main__":
    _main()



import numpy as np
import numpy.typing as npt
import torch
import torch.nn.functional as F
from torch import nn


class AdaptiveAvgPool2d(nn.Module):
    r"""
    Applies a 2D adaptive average pooling over an input signal composed of several input planes.

    The output is of size :math:`H \times W`, for any input size. The number of output
    features is equal to the number of input planes.

    .. note::
        This class is a re-implementation of ``torch.nn.AdaptiveAvgPool2d`` that can be
        exported to ONNX and serves as a drop-in replacement for torch's version.

    Args:
        output_size: the target output size of the image of the form :math:`H \times W`.
            Can be a tuple ``(H, W)`` or a single ``H`` for a square image :math:`H \times
            H`. ``H`` and ``W`` can be either a ``int``, or ``None`` which means the size
            will be the same as that of the input.

    """  # noqa: E501

    def __init__(self, output_size: int | tuple[int | None, int | None]):
        """Create the pool."""
        super().__init__()

        self._validate_output_size(output_size)

        if isinstance(output_size, int):
            output_size = (output_size, output_size)

        self._output_size = output_size

    def _validate_output_size(
        self, output_size: int | tuple[int | None, int | None]
    ) -> None:
        if not isinstance(output_size, int | tuple):
            raise RuntimeError("error: output_size must be an int or a tuple")
        if isinstance(output_size, tuple):
            if len(output_size) != 2:
                raise RuntimeError(
                    f"error: output_size must have size 2 but received {len(output_size)}"
                )
            if any(not isinstance(o, int | None) for o in output_size):
                raise RuntimeError("error: expected a tuple of int or None")

    def _get_output_size(self, x: torch.Tensor) -> npt.NDArray:
        x_size = np.array(x.shape[-2:])
        if self._output_size is None:
            return x_size

        # At this point, the output size should just be a tuple
        assert isinstance(self._output_size, tuple)
        x_size[0] = (
            self._output_size[0] if self._output_size[0] is not None else x_size[0]
        )
        x_size[1] = (
            self._output_size[1] if self._output_size[1] is not None else x_size[1]
        )
        return x_size

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Evaluate the adaptive average pool on the input.

        Args:
            x: the input tensor.

        Returns:
            The pooled tensor.
        """
        output_size = self._get_output_size(x)
        stride_size = np.floor(np.array(x.shape[-2:]) / output_size).astype(np.int32)
        kernel_size = np.array(x.shape[-2:]) - (output_size - 1) * stride_size  # type: ignore[operator]
        x = F.avg_pool2d(x, kernel_size=list(kernel_size), stride=list(stride_size))
        return x



import typing

from torch import optim

from helios import core


def _register_default_optimizers(registry: core.Registry) -> None:
    """
    Register the Torch default optimizers.

    List was obtained from
    https://pytorch.org/docs/stable/optim.html#algorithms

    Args:
        registry: the optimizer registry.
    """
    registry.register(optim.Adadelta)
    registry.register(optim.Adagrad)
    registry.register(optim.Adam)
    registry.register(optim.AdamW)
    registry.register(optim.SparseAdam)
    registry.register(optim.Adamax)
    registry.register(optim.ASGD)
    registry.register(optim.LBFGS)
    registry.register(optim.NAdam)
    registry.register(optim.RAdam)
    registry.register(optim.RMSprop)
    registry.register(optim.Rprop)
    registry.register(optim.SGD)


OPTIMIZER_REGISTRY = core.Registry("optimizer")
"""
Global instance of the registry for optimizers.

By default, the registry contains the following optimizers:

.. list-table:: Optimizers
    :header-rows: 1

    * - Optimizer
      - Name
    * - ``torch.optim.Adadelta``
      - Adadelta
    * - ``torch.optim.Adagrad``
      - Adagrad
    * - ``torch.optim.Adam``
      - Adam
    * - ``torch.optim.AdamW``
      - AdamW
    * - ``torch.optim.SparseAdam``
      - SparseAdam
    * - ``torch.optim.Adamax``
      - Adamax
    * - ``torch.optim.ASGD``
      - ASGD
    * - ``torch.optim.LBFGS``
      - LBFGS
    * - ``torch.optim.NAdam``
      - NAdam
    * - ``torch.optim.RAdam``
      - RAdam
    * - ``torch.optim.RMSprop``
      - RMSprop
    * - ``torch.optim.Rprop``
      - Rprop
    * - ``torch.optim.SGD``
      - SGD

Example:
    .. code-block:: python

        import helios.optim as hlo

        # This automatically registers your optimizer.
        @hlo.OPTIMIZER_REGISTRY.register
        class MyOptimizer:
            ...

        # Alternatively you can manually register a optimizer like this:
        hlo.OPTIMIZER_REGISTRY.register(MyOptimizer)
"""
_register_default_optimizers(OPTIMIZER_REGISTRY)


def create_optimizer(
    type_name: str, *args: typing.Any, **kwargs: typing.Any
) -> optim.Optimizer:
    """
    Create the optimizer for the given type.

    Args:
        type_name: the type of the optimizer to create.
        args: positional arguments to pass into the optimizer.
        kwargs: keyword arguments to pass into the optimizer.
    Returns:
        The optimizer.
    """
    return OPTIMIZER_REGISTRY.get(type_name)(*args, **kwargs)



"""
Core package for Helios.

Holds all of the core functions and classes used throughout the package.
"""

from . import cuda, distributed, logging, rng
from .utils import (
    AverageTimer,
    ChdirContext,
    Registry,
    convert_to_list,
    get_env_info_str,
    get_from_optional,
    safe_torch_load,
    update_all_registries,
)

__all__ = [
    "cuda",
    "distributed",
    "logging",
    "rng",
    "AverageTimer",
    "ChdirContext",
    "Registry",
    "convert_to_list",
    "get_env_info_str",
    "get_from_optional",
    "safe_torch_load",
    "update_all_registries",
]
